/**
Very simple web server.
This version serves only one client at a time.

Orion Sky Lawlor, olawlor@acm.org, 2005/2/8 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "osl/socket.h"
#include "osl/socket.cpp" /* include body for easy linking */

/* Convert integer to string */
std::string int_to_string(int i) {
	char buf[100]; sprintf(buf,"%d",i); return buf;
}

void kill_client(const char *why,const char *what) {
	printf("Client sent bogus request: %s '%s'\n",why,what);
	exit(1); /*<- In a real server, this would be "give up and go to the next client... */
}

void serve_client(SERVER_SOCKET server_fd)
{
// Connect with the client
	skt_ip_t client_ip; unsigned int client_port;
	SOCKET s=skt_accept(server_fd,&client_ip,&client_port);
	char client_ip_name[200];
	printf("Connection received from %s:%d\n",
		skt_print_ip(client_ip_name,client_ip),client_port);
	
// Receive the client's request up to the first newline.
	//  This looks like "GET /some/url HTTP/1.1"
	std::string req_prot=skt_recv_string(s," \r\n"); /* <- Aniel's idea: use skt_recv_str instead of breaking up line into pieces ourselves */
	std::string req_url=skt_recv_string(s," \r\n");
	std::string req_ver=skt_recv_line(s);
	
	/* HTTP request options */
	std::string options="";
	while (1) {
		std::string in=skt_recv_line(s);
		if (in.size()==0) break;
		else options+=in+"\n";
	}
	printf("Client request URL: '%s'\n",req_url.c_str());
	
	/* Protocol checking */
	if (req_prot!="GET") kill_client("bad HTTP request",req_prot.c_str());
	if (strncmp(req_url.c_str(),"HTTP/1.",7)==0) kill_client("bad HTTP version (only 1.x supported)",req_ver.c_str());
	
	if (strncmp(req_url.c_str(),"http://",7)==0) 
	{ /* strip off http and host name, leaving bare path URL */
		int i=7; /* pointing at start of host name */
		while (i<req_url.size() && req_url[i]!='/') i++;
		if (i>=req_url.size() || req_url[i]!='/') kill_client("No path present in request!",req_url.c_str());
		/* chop out substring of req_url starting at i (the path) */
		req_url=req_url.substr(i);
		printf("Chopped out hostname.  URL now: '%s'\n",req_url.c_str());
	}
	
	/* Security checking */
	if (strstr(req_url.c_str(),".."))  /* imagine "../../../../etc/passwd" */
		kill_client("Trying to access enclosing directory",req_url.c_str());
	if (strstr(req_url.c_str(),":"))  /* imagine "C:\boot.ini" */
		kill_client("Trying to access path containing a colon",req_url.c_str());
	if (req_url.size()<1 || req_url.size()>10000) 
		kill_client("Bogus request URL length",req_url.c_str());
	if (req_url[0]!='/')
		kill_client("Request URL must start with slash",req_url.c_str());
	
// Respond to the client's request:
	if (req_url=="/cgi-bin/shutdown.cgi") {
		printf("Exiting webserver by client CGI request\n");
		exit(0);
	}
	
	/* Try to open the file */
	std::string req_file="public_html"+req_url; /* URL must start with slash */
	FILE *f=fopen(req_file.c_str(),"rb");
	if (f==NULL) { /* <- Couldn't open file.  Return a 404. */
		printf("File not found.  Returning 404.\n");
		const char *err=
		"HTTP/1.0 404 OK\r\n"
		"Content-Type: text/html\r\n"
		"Content-Length: 37\r\n"
		"\r\n"
		"<html><body>Not found</body></html>\r\n";
		skt_sendN(s,err,strlen(err));
		return;
	}
	else { /* file exists.  Send it! */
		// Find file length
		fseek(f,0,SEEK_END);
		int filelen=ftell(f);
		fseek(f,0,SEEK_SET); /* back to start of file */
		printf("File opened as '%s', length %d bytes\n",req_file.c_str(),filelen);

		// Prepare HTTP headers
		std::string headers= /* HTTP headers, less Content-Length */
		        "HTTP/1.0 200 OK\r\n"
		        "Content-Type: text/html\r\n" /* <- FIXME: support other types here */
		        "Content-Length: "+int_to_string(filelen)+"\r\n"
		        "\r\n";
		skt_sendN(s,headers.c_str(),headers.size());
		
		// Send chunks of file until done
		int readlen;
		do {
			enum {buflen=4096};
			char buf[buflen];
			readlen=fread(buf,1,buflen,f);
			skt_sendN(s,buf,readlen);
			printf("Sent off %d bytes\n",readlen);
		} while (readlen>0);
	}
	
	// Close the client socket
	skt_close(s);
}

int main(int argc,char *argv[])
{
	unsigned int port=1234;
	SERVER_SOCKET server_fd;
	skt_init();
	server_fd=skt_server(&port);
	while (1) {
		printf("\nWaiting for connect on port %d\n",port);
		// Wait for a client to connect
		skt_select1(server_fd,0);
		// Serve this client
		serve_client(server_fd);
	}
	return 0;
}

